from flask import request, Response

from flipbook import args, RELATIVE_DIRECTORY_TO_DATA_FILES_LIST, FORM_RESPONSES, FORM_SCHEMA_COLUMNS, \
    RELATIVE_DIRECTORY_TO_METADATA, METADATA_COLUMNS, EXTRA_DATA_IN_FORM_RESPONSES_TABLE, \
    EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE, get_static_data_page_url
from flipbook.utils import load_jinja_template, get_data_page_url

MAIN_LIST_TEMPLATE = None


def main_list_handler(is_static_website=False):
    global MAIN_LIST_TEMPLATE
    if MAIN_LIST_TEMPLATE is None or args.dev_mode:
        MAIN_LIST_TEMPLATE = load_jinja_template("main_list")

    if args.verbose:
        print(f"main_list_handler recieved {request.url}")

    data_files_list = [
        (page_number + 1, relative_directory, data_file_types_and_paths)
        for page_number, (relative_directory, data_file_types_and_paths) in enumerate(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)
    ]

    # combine metadata from 2 potential sources: RELATIVE_DIRECTORY_TO_METADATA and EXTRA_DATA_IN_FORM_RESPONSES_TABLE
    metadata_columns = []
    metadata_dict = {}
    if not args.hide_metadata_on_home_page:
        metadata_columns = METADATA_COLUMNS
        if not is_static_website:
            metadata_columns += EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE
        for relative_dir in list(RELATIVE_DIRECTORY_TO_METADATA.keys()) + list(EXTRA_DATA_IN_FORM_RESPONSES_TABLE.keys()):
            metadata_dict[relative_dir] = dict(RELATIVE_DIRECTORY_TO_METADATA.get(relative_dir, {}))
            metadata_dict[relative_dir].update(dict(EXTRA_DATA_IN_FORM_RESPONSES_TABLE.get(relative_dir, {})))

    html = MAIN_LIST_TEMPLATE.render(
        data_files_list=data_files_list,
        get_data_page_url=get_data_page_url if not is_static_website else get_static_data_page_url,
        form_column_names=FORM_SCHEMA_COLUMNS if not is_static_website else [],
        form_responses_dict=FORM_RESPONSES,
        metadata_column_names=metadata_columns,
        metadata_dict=metadata_dict,
        form_responses_table_path=args.form_responses_table,
        is_static_website=is_static_website,
    )

    return Response(html, mimetype='text/html')





import os
from flask import request, Response
from pprint import pprint, pformat
from flipbook import args, RELATIVE_DIRECTORY_TO_DATA_FILES_LIST, FORM_SCHEMA, FORM_RESPONSES, \
    RELATIVE_DIRECTORY_TO_METADATA, FORM_RADIO_BUTTON_KEYBOARD_SHORTCUTS, EXTRA_DATA_IN_FORM_RESPONSES_TABLE, \
    get_static_data_page_url
from flipbook.utils import load_jinja_template, get_data_page_url, CONTENT_HTML_FILE_TYPE, \
    IMAGE_FILE_TYPE

DATA_PAGE_TEMPLATE = None


def data_page_handler(is_static_website=False):
    global DATA_PAGE_TEMPLATE
    if DATA_PAGE_TEMPLATE is None or args.dev_mode:
        DATA_PAGE_TEMPLATE = load_jinja_template("data_page")

    if args.verbose:
        print(f"data_page_handler received {request.url}")

    params = {}
    if dict(request.args):
        params.update(dict(request.args))
    if dict(request.values):
        params.update(dict(request.values))

    json_args = request.get_json(force=True, silent=True)
    if 'i' not in params:
        params.update(json_args or {})

    if args.verbose > 1:
        print(f"data_page_handler request.data: {pformat(request.data)}")
        print(f"data_page_handler request.form: {bool(request.form)} {pformat(request.form)}")
        print(f"data_page_handler request.args: {bool(request.args)} {pformat(request.args)}")
        print(f"data_page_handler request.values: {bool(request.values)} {pformat(request.values)}")
        print(f"data_page_handler request.get_json(..): {bool(json_args)} {pformat(json_args)}")
        print(f"data_page_handler request.__dict__: {pformat(request.__dict__)}")
        print(f"data_page_handler params: {params}")

    i = None
    relative_dir = params.get("path")
    if relative_dir:
        # override i
        for idx, (known_relative_dir, _) in enumerate(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST):
            if relative_dir == known_relative_dir:
                i = idx + 1
                break
        else:
            print(f"ERROR: path param '{relative_dir}' not recognized. Falling back on using i param.")

    if i is None:
        i = params.get("i")

    try:
        if isinstance(i, list):
            i = int(i[0])
        else:
            i = int(i)
    except (ValueError, TypeError) as e:
        print(f"ERROR: unable to parse parameter i: '{i}': {type(e).__name__} {e}. Setting i = 1.")
        i = 1

    last = params.get("last", i)
    try:
        if isinstance(i, list):
            last = int(last[0])
        else:
            last = int(last)
    except (ValueError, TypeError) as e:
        print(f"ERROR: unable to parse parameter 'last': '{last}': {type(e).__name__} {e}. Setting last = {i}.")
        last = i

    if last < 1:
        print(f"ERROR: parameter 'last' (= {last}) is less than 1. Resetting it to 1.")
        last = 1

    if i < 1:
        print(f"ERROR: parameter i (= {i}) is less than 1. Resetting it to 1.")
        i = 1

    if i > len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST):
        print(f"ERROR: parameter i (= {i}) is greater than the # of pages "
              f"(= {len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)}). "
              f"Resetting it to {len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)}.")
        i = len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)

    relative_dir, data_file_types_and_paths = RELATIVE_DIRECTORY_TO_DATA_FILES_LIST[i - 1]

    image_file_paths = []
    for data_file_type, data_file_path in data_file_types_and_paths:
        if data_file_type == IMAGE_FILE_TYPE:
            image_file_paths.append(data_file_path)

    metadata_json_dict = RELATIVE_DIRECTORY_TO_METADATA.get(relative_dir, {})
    metadata_json_dict.update(EXTRA_DATA_IN_FORM_RESPONSES_TABLE.get(relative_dir, {}))

    content_html_strings = []
    for data_file_type, data_file_path in data_file_types_and_paths:
        if data_file_type != CONTENT_HTML_FILE_TYPE:
            continue
        with open(os.path.join(args.directory, data_file_path), "rt") as f:
            content_string = f.read()
            content_html_strings.append((data_file_path, content_string))

    if args.verbose:
        print(f"data_page_handler returning i={i}, last={last}, relative_directory={relative_dir}, "
              f"{len(image_file_paths)} image_file_paths, {len(metadata_json_dict)} records in metadata_json_dict, "
              f"{len(content_html_strings)} content_html_strings")

    html = DATA_PAGE_TEMPLATE.render(
        i=i,
        last=last,
        relative_directory=relative_dir,
        image_file_paths=image_file_paths,
        metadata_json_dict=metadata_json_dict,
        content_html_strings=content_html_strings,
        get_data_page_url=get_data_page_url if not is_static_website else get_static_data_page_url,
        form_schema=FORM_SCHEMA if not is_static_website else [],
        form_radio_button_keyboard_shortcuts=FORM_RADIO_BUTTON_KEYBOARD_SHORTCUTS if not is_static_website else {},
        form_responses=FORM_RESPONSES.get(relative_dir, {}),
        is_static_website=is_static_website,
    )

    return Response(html, mimetype='text/html')




import argparse
import collections
import pandas as pd

EXPECTED_COLUMNS = {"Path", "Verdict", "Confidence"}
HIGH_CONFIDENCE = "high"

DISCORDANT_VERDICT_COLUMN = "Discordant Verdict"
DISCORDANT_SCORE_COLUMN = "Discordance Score"
DISCORDANT_TEXT_COLUMN = "Discordance Text"


def parse_args():
    p = argparse.ArgumentParser(description="Check concordance between two form response tables created by 2 users "
        "reviewing the same images. Rows in the two tables are matched by their same Path column.")
    p.add_argument("-s1", "--suffix1", help="Suffix to append to column names from table1", default="1")
    p.add_argument("-s2", "--suffix2", help="Suffix to append to column names from table2", default="2")
    p.add_argument("-o", "--output-table", help="Path of output .tsv or .xls", default="combined.tsv")
    p.add_argument("table1", help="Path of form response table .tsv")
    p.add_argument("table2", help="Path of form response table .tsv")
    args = p.parse_args()

    try:
        if args.table1.endswith(".xls") or args.table1.endswith(".xlsx"):
            df1 = pd.read_excel(args.table1, engine="openpyxl")
        else:
            df1 = pd.read_table(args.table1)
    except Exception as e:
        p.error(f"Error parsing {args.table1}: {e}")

    try:
        if args.table2.endswith(".xls") or args.table2.endswith(".xlsx"):
            df2 = pd.read_excel(args.table2, engine="openpyxl")
        else:
            df2 = pd.read_table(args.table2)
    except Exception as e:
        p.error(f"Error parsing {args.table2}: {e}")

    if EXPECTED_COLUMNS - set(df1.columns):
        p.error(f"{args.table1} is missing columns: " + ", ".join(EXPECTED_COLUMNS - set(df1.columns)))

    if EXPECTED_COLUMNS - set(df2.columns):
        p.error(f"{args.table2} is missing columns: " + ", ".join(EXPECTED_COLUMNS - set(df2.columns)))

    if len(df1) == 0:
        p.error(f"{args.table1} is empty")

    if len(df2) == 0:
        p.error(f"{args.table2} is empty")

    df1 = df1.set_index("Path").fillna("")
    df2 = df2.set_index("Path").fillna("")

    if len(set(df1.index) & set(df2.index)) == 0:
        p.error(f"{args.table1} Path column values have 0 overlap with {args.table2} Path column values. Tables can only "
                f"be combined if they have the same Paths.")

    return args, df1, df2


def compute_discordance_columns_func(suffix1, suffix2):
    verdict_column1 = f"Verdict_{suffix1}"
    verdict_column2 = f"Verdict_{suffix2}"
    confidence_column1 = f"Confidence_{suffix1}"
    confidence_column2 = f"Confidence_{suffix2}"

    def compute_discordance_columns(row):
        if not row[verdict_column1] or not row[verdict_column2]:
            return row

        if row[verdict_column1] == row[verdict_column2]:
            # same verdict
            row[DISCORDANT_VERDICT_COLUMN] = 0
            row[DISCORDANT_SCORE_COLUMN] = 0
            row[DISCORDANT_TEXT_COLUMN] = "same verdict"
            if row[confidence_column1] and row[confidence_column2]:
                if row[confidence_column1] == HIGH_CONFIDENCE and row[confidence_column2] == HIGH_CONFIDENCE:
                    row[DISCORDANT_SCORE_COLUMN] = 0
                    row[DISCORDANT_TEXT_COLUMN] = "same verdict, both high confidence"
                elif row[confidence_column1] == HIGH_CONFIDENCE or row[confidence_column2] == HIGH_CONFIDENCE:
                    row[DISCORDANT_SCORE_COLUMN] = 1
                    row[DISCORDANT_TEXT_COLUMN] = "same verdict, one high confidence"
                else:
                    row[DISCORDANT_SCORE_COLUMN] = 0
                    row[DISCORDANT_TEXT_COLUMN] = "same verdict, zero high confidence"
        else:
            # different verdicts
            row[DISCORDANT_VERDICT_COLUMN] = 1
            row[DISCORDANT_SCORE_COLUMN] = 2
            row[DISCORDANT_TEXT_COLUMN] = "different verdict"
            if row[confidence_column1] or row[confidence_column2]:
                if row[confidence_column1] == HIGH_CONFIDENCE and row[confidence_column2] == HIGH_CONFIDENCE:
                    row[DISCORDANT_SCORE_COLUMN] = 4
                    row[DISCORDANT_TEXT_COLUMN] = "different verdict, both high confidence"
                elif row[confidence_column1] == HIGH_CONFIDENCE or row[confidence_column2] == HIGH_CONFIDENCE:
                    row[DISCORDANT_SCORE_COLUMN] = 3
                    row[DISCORDANT_TEXT_COLUMN] = "different verdict, one high confidence"
                else:
                    row[DISCORDANT_SCORE_COLUMN] = 2
                    row[DISCORDANT_TEXT_COLUMN] = "different verdict, zero high confidence"

        return row

    return compute_discordance_columns


def main():
    args, df1, df2 = parse_args()
    df_joined = df1.join(df2, lsuffix=f"_{args.suffix1}", rsuffix=f"_{args.suffix2}", how="outer").reset_index()
    df_joined = df_joined.fillna("")

    #  print stats about input tables
    def get_counts_string(df, column, label="", sep=", "):
        return sep.join(
            [f"{count:2d} {label} {key}" for key, count in sorted(collections.Counter(df[column].fillna("<empty>")).items())])

    print("-"*20)
    df1_verdicts_counter = get_counts_string(df1, "Verdict")
    df1_num_verdicts = sum(df1['Verdict'].str.len() > 0)
    df1_num_high_confidence = collections.Counter(df1['Confidence']).get(HIGH_CONFIDENCE, 0)
    df1_high_confidence_fraction = df1_num_high_confidence / df1_num_verdicts

    df2_verdicts_counter = get_counts_string(df2, "Verdict")
    df2_num_verdicts = sum(df2['Verdict'].str.len() > 0)
    df2_num_high_confidence = collections.Counter(df2['Confidence']).get(HIGH_CONFIDENCE, 0)
    df2_high_confidence_fraction = df2_num_high_confidence / df2_num_verdicts

    print(f"{args.table2}:  {df1_num_verdicts} verdicts. {df1_verdicts_counter}. High confidence for {100*df1_high_confidence_fraction:4.1f}% of them")
    print(f"{args.table1}:  {df2_num_verdicts} verdicts. {df2_verdicts_counter}. High confidence for {100*df2_high_confidence_fraction:4.1f}% of them")

    #  compute concordance
    df_joined = df_joined.apply(compute_discordance_columns_func(args.suffix1, args.suffix2), axis=1)

    # print concordance stats
    print("-"*20)
    num_discordant_verdicts = sum(df_joined[DISCORDANT_VERDICT_COLUMN])
    print(f"{num_discordant_verdicts} out of {len(df_joined)} ({100*num_discordant_verdicts/len(df_joined):0.1f}%) of "
          f"verdicts differed between the two tables")
    print(f"\nDiscordance score = {sum(df_joined[DISCORDANT_SCORE_COLUMN])}:")
    print(get_counts_string(df_joined, DISCORDANT_TEXT_COLUMN, label="review comparisons:", sep="\n"))

    # write outtable
    print("-"*20)
    verdict_column1 = f"Verdict_{args.suffix1}"
    confidence_column1 = f"Confidence_{args.suffix1}"
    df_joined.sort_values(
        [DISCORDANT_SCORE_COLUMN, DISCORDANT_TEXT_COLUMN, verdict_column1, confidence_column1],
        ascending=False, inplace=True)
    if args.output_table.endswith(".xls") or args.output_table.endswith(".xlsx"):
        df_joined.to_excel(args.output_table, header=True, index=False)
    else:
        df_joined.to_csv(args.output_table, header=True, sep="\t", index=False)
    print(f"Wrote {len(df_joined)} rows to {args.output_table}")


if __name__ == "__main__":
    main()



import collections
from jinja2 import Template
import json
import os
import pkg_resources
import sys
from wcmatch import glob

IMAGE_FILE_TYPE = "image"
METADATA_JSON_FILE_TYPE = "metadata_json"
CONTENT_HTML_FILE_TYPE = "content_html"

METADATA_JSON_FILE_SUFFIX = "flipbook_metadata.json"
CONTENT_HTML_FILE_SUFFIX = "flipbook_content.html"


def get_relative_directory_to_data_files_list(
        top_level_dir,
        keywords_to_include,
        keywords_to_exclude,
        suffixes=("svg", "png", "jpeg", "jpg", "gif", "webp", CONTENT_HTML_FILE_SUFFIX, METADATA_JSON_FILE_SUFFIX),
        verbose=False):
    data_file_paths = []

    print(f"Looking for " + ", ".join(suffixes[:-1]) + f", and {suffixes[-1]} files in {top_level_dir}")
    if keywords_to_include or keywords_to_exclude:
        sys.stdout.write("Keeping only file paths that ")
        if keywords_to_include:
            sys.stdout.write("contain " + " or ".join([f'"{k}"' for k in keywords_to_include]))
        if keywords_to_include and keywords_to_exclude:
            sys.stdout.write(" and ")
        if keywords_to_exclude:
            sys.stdout.write("don't contain " + " or ".join([f'"{k}"' for k in keywords_to_exclude]))
        sys.stdout.write("\n")

    glob_string = "|".join([f"{top_level_dir}/**/*{suffix}" for suffix in suffixes]).replace("\\", "/")
    matching_paths = glob.glob(glob_string, flags=glob.GLOBSTAR|glob.SPLIT)
    data_file_paths += matching_paths

    # group data files by their directory
    subdirectories_list = []
    data_file_counter_by_suffix = collections.defaultdict(int)
    relative_directory_to_data_files = collections.defaultdict(list)
    excluded_keyword_to_matching_paths = collections.defaultdict(list)
    for data_file_path in data_file_paths:
        data_file_path = os.path.realpath(data_file_path)
        for suffix in suffixes:
            if data_file_path.endswith(suffix):
                data_file_suffix = suffix
                break
        else:
            raise Exception(f"Unexpected file suffix: {data_file_path}")

        data_file_counter_by_suffix[data_file_suffix] += 1
        relative_data_file_path = os.path.relpath(data_file_path, top_level_dir)

        if keywords_to_include:
            included_keyword_matches = [k for k in keywords_to_include if k in relative_data_file_path] if keywords_to_include else []
            if not included_keyword_matches:
                if verbose:
                    print(f"Skipping {data_file_suffix} file: {relative_data_file_path} - it doesn't contain any --include keyword.")
                continue

        excluded_keyword_matches = [k for k in keywords_to_exclude if k in relative_data_file_path] if keywords_to_exclude else []
        if excluded_keyword_matches:
            excluded_keyword_to_matching_paths[excluded_keyword_matches[0]].append(relative_data_file_path)
            if verbose:
                print(f"Skipping {data_file_suffix} file: {relative_data_file_path} - it contains excluded keyword: '{excluded_keyword_matches[0]}'")
            continue
        else:
            if verbose:
                print(f"{data_file_suffix} file: {relative_data_file_path}")

        key = os.path.dirname(relative_data_file_path)
        if key and key != ".":
            subdirectories_list.append(key)
        else:
            key = relative_data_file_path

        if data_file_suffix == METADATA_JSON_FILE_SUFFIX:
            data_file_type = METADATA_JSON_FILE_TYPE
        elif data_file_suffix == CONTENT_HTML_FILE_SUFFIX:
            data_file_type = CONTENT_HTML_FILE_TYPE
        else:
            data_file_type = IMAGE_FILE_TYPE

        relative_directory_to_data_files[key].append((data_file_type, relative_data_file_path))

    data_file_counter_string = ", ".join([
        ("" if i < len(data_file_counter_by_suffix) - 1 or len(data_file_counter_by_suffix) < 2 else "and ") +
        f"{counter} {suffix} file" +
        ("s" if counter != 1 else "")
        for i, (suffix, counter) in enumerate(data_file_counter_by_suffix.items())
    ]) if data_file_counter_by_suffix else "0 matching files"
    print(f"Found {data_file_counter_string}" + (
        f" in {len(subdirectories_list)} subdirectories" if len(subdirectories_list) > 1 else ""
    ))

    for excluded_keyword, matching_paths in excluded_keyword_to_matching_paths.items():
        print(f"Skipped {len(matching_paths)} paths which contained excluded keyword: '{excluded_keyword}'")

    relative_directory_to_data_files_list = list(sorted(relative_directory_to_data_files.items()))

    return relative_directory_to_data_files_list


def get_relative_directory_to_metadata(top_level_dir, relative_directory_to_data_files_list, verbose=False):
    metadata_columns = collections.OrderedDict()
    relative_directory_to_metadata = {}
    # parse the flipbook_metadata.json file from each directory
    for relative_dir, data_files_list in relative_directory_to_data_files_list:
        for data_file_type, data_file_path in data_files_list:
            if data_file_type != METADATA_JSON_FILE_TYPE:
                continue

            metadata_json_path = os.path.join(top_level_dir, data_file_path)
            try:
                with open(metadata_json_path, "rt") as f:
                    metadata_json = json.load(f)
            except Exception as e:
                print(f"Unable to parse {metadata_json_path}: {e}")
                continue

            if not isinstance(metadata_json, dict):
                print(f"WARNING: {metadata_json_path} doesn't contain a dictionary. Skipping...")
                continue

            if verbose:
                print(f"Parsed {len(metadata_json)} metadata entries from {metadata_json_path}")

            relative_directory_to_metadata[relative_dir] = metadata_json

            for key in metadata_json:
                metadata_columns[key] = None

    if relative_directory_to_metadata:
        print(f"Parsed {len(relative_directory_to_metadata)} {METADATA_JSON_FILE_SUFFIX} files" + (
            f" with columns: {', '.join(metadata_columns)}" if metadata_columns else ""))

    return list(metadata_columns.keys()), relative_directory_to_metadata


def is_excel_table(path):
    return any(path.endswith(suffix) for suffix in ("xls", "xlsx"))


def get_data_page_url(page_number, last):
    return f"/page?last={last}&i={page_number}"


def load_jinja_template(name):
    return Template(pkg_resources.resource_stream("flipbook" , f"templates/{name}.html").read().decode('UTF-8'))



import json
from flask import request, Response
import pandas as pd

from flipbook import args, FORM_SCHEMA, FORM_RESPONSES, FORM_SCHEMA_COLUMNS, PATH_COLUMN, \
    METADATA_COLUMNS, RELATIVE_DIRECTORY_TO_METADATA, EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE, \
    EXTRA_DATA_IN_FORM_RESPONSES_TABLE


def error_response(message, status=400):
    return Response(json.dumps({"error": message}), status=status, mimetype='application/json')


def save_form_handler():
    if not FORM_SCHEMA:
        return error_response("Server state error: form table not initialized", status=500)

    # check params
    params = request.form
    if 'relative_directory' not in params:
        return error_response("'relative_directory' not provided")

    # transfer values to FORM_RESPONSES
    for form_schema_row in [{'name': 'relative_directory', 'columnName': PATH_COLUMN}] + FORM_SCHEMA:
        value = params.get(form_schema_row['name'])
        if value is None:
            continue
        if params['relative_directory'] not in FORM_RESPONSES:
            FORM_RESPONSES[params['relative_directory']] = {}
            if args.verbose:
                print(f"Adding {params['relative_directory']} row to responses")

        FORM_RESPONSES[params['relative_directory']][form_schema_row['columnName']] = value
        if args.verbose:
            print(f"Setting {params['relative_directory']} {form_schema_row['columnName']} = {value}")

    # transfer metadata values to FORM_RESPONSES
    output_table_rows = []
    output_table_columns = [PATH_COLUMN] + FORM_SCHEMA_COLUMNS + EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE
    if args.add_metadata_to_form_responses_table:
        output_table_columns += METADATA_COLUMNS

    for relative_dir in FORM_RESPONSES:  # set(RELATIVE_DIRECTORY_TO_METADATA.keys()) |
        output_dict = {
            PATH_COLUMN: relative_dir,
        }
        if args.add_metadata_to_form_responses_table:
            output_dict.update(RELATIVE_DIRECTORY_TO_METADATA.get(relative_dir, {}))
        output_dict.update(FORM_RESPONSES.get(relative_dir, {}))
        output_dict.update(EXTRA_DATA_IN_FORM_RESPONSES_TABLE.get(relative_dir, {}))
        output_table_rows.append(output_dict)

    # write table to file
    # NOTE: This is not thread-safe and assumes a single-threaded server. For multi-threaded
    # or multi-process servers like gunicorn, this will need to be replaced with a sqlite or redis backend.
    df = pd.DataFrame(output_table_rows, columns=output_table_columns).fillna('')
    print(f"Saving {len(df)} rows to {args.form_responses_table}")
    if args.form_responses_table_is_excel:
        df.to_excel(args.form_responses_table)
    else:
        df.to_csv(args.form_responses_table, sep="\t", header=True, index=False)

    return Response(json.dumps({"success": True}), status=200, mimetype='application/json')




import collections
import configargparse
from flask import Flask, Response, send_from_directory
from flask_cors import CORS
import jinja2
import json
import os
import pandas as pd
import pkg_resources
import pkgutil
import re
import requests
import shutil
import sys

from flipbook.utils import get_relative_directory_to_data_files_list, get_relative_directory_to_metadata, \
    is_excel_table, get_data_page_url, METADATA_JSON_FILE_TYPE, CONTENT_HTML_FILE_TYPE

PATH_COLUMN = 'Path'
WEBSITE_DIR = "flipbook_html"

p = configargparse.ArgumentParser(
    formatter_class=configargparse.DefaultsFormatter,
    add_config_file_help=True,
    add_env_var_help=True,
    config_file_parser_class=configargparse.YAMLConfigFileParser,
    default_config_files=["~/.flipbook_config"],
    args_for_writing_out_config_file=["--save-current-options-to-config-file"],
)
p.add_argument("-i", "--include", action="append", help="Only include files whose path contains this keyword")
p.add_argument("-x", "--exclude", action="append", help="Skip files whose path contains this keyword. If both "
               " --include and --exclude are specified, --exclude takes precedence over --include", default=[WEBSITE_DIR])
p.add_argument("-t", "--form-responses-table", default="flipbook_form_responses.tsv",
               help="The .tsv or .xls path where form responses are saved. If the file already exists,"
                    "it will be parsed for previous form responses and then updated as the user fills in the form(s)."
                    "If the file doesn't exist, it will be created after the 1st form response.")
p.add_argument("-m", "--metadata-table", default="flipbook_metadata.tsv",
               help="The .tsv or .xls path containing metadata to show on data pages. There are two optional ways "
                    "to add metadata to the data pages. The 1st way is to put a 'flipbook_metadata.json' file "
                    "inside a directory that contains images or data files (in which case any key-value pairs from "
                    "the json file will be shown at the top of the data page that displays those images). "
                    "The other way is to specify this table, which needs to have a 'Path' column with relative "
                    "directory paths that contain images and data files. The data page corresponding to those "
                    "directory paths will then display values from the other columns in this table. If both this table "
                    "and 'flipbook_metadata.json' files are found, the values from this table will override values in "
                    "the 'flipbook_metadata.json' files.")
p.add_argument("-j", "--form-schema-json", help="Path of .json file containing a custom form schema. For the expected format "
               "see https://github.com/broadinstitute/flipbook/tree/main/form_schema_examples")
p.add_argument("-s", "--sort-by", action="append", help="Order pages by metadata column(s)")
p.add_argument("--hide-metadata-on-home-page", action="store_true", help="Don't show metadata columns in the "
               "home page table")
p.add_argument("--add-metadata-to-form-responses-table", action="store_true", help="Also write metadata columns to the "
               "form responses table when saving users' form responses")

p.add_argument("--generate-static-website", action="store_true", help="Instead of starting a web server, this option "
               "causes FlipBook to write out a set of static html pages for all the images it finds and then exit. "
               "The generated pages can then be viewed in a browser, uploaded to some other web server (such as "
               "GitHub Pages, embedded in another existing website, etc. The generated web pages are identical to "
               "the standard FlipBook user interface except they don't contain the forms for entering responses about "
               "each image - and so just allow flipping through the images.")

#p.add_argument("-c", "--config-file", help="Path of yaml config file", env_var="FLIPBOOK_CONFIG_FILE")
p.add_argument("-v", "--verbose", action='count', default=0, help="Print more info")
p.add_argument("--host", default="127.0.0.1", env_var="HOST", help="Listen for connections on this hostname or IP")
p.add_argument("-p", "--port", default="8080", env_var="PORT", type=int, help="Listen for connections on this port")
p.add_argument("--dev-mode", action="store_true", env_var="DEV", help="Run server in developer mode so it reloads "
               "html templates and source code if they're changed")
p.add_argument("directory", default=".", nargs="?", help="Top-level directory to search for images and data files")
args = p.parse_args()

if args.verbose > 1:
    p.print_values()

if not os.path.isdir(args.directory):
    p.error(f"{args.directory} directory not found")

args.directory = os.path.realpath(args.directory)


def parse_table(path):
    if not os.path.isfile(path):
        raise ValueError(f"{path} not found")

    try:
        if is_excel_table(path):
            df = pd.read_excel(path, engine="openpyxl")
        else:
            df = pd.read_table(path)
    except Exception as e:
        raise ValueError(f"Unable to parse {path}: {e}")

        # validate table contents
    if PATH_COLUMN not in df.columns:
        raise ValueError(f"{path} must have a column named '{PATH_COLUMN}'")

    df.set_index(PATH_COLUMN, inplace=True, drop=False)

    df = df.fillna('')
    print(f"Parsed {len(df)} rows from {path}")

    return df


# search directory for images and data files
RELATIVE_DIRECTORY_TO_DATA_FILES_LIST = get_relative_directory_to_data_files_list(
    args.directory,
    args.include,
    args.exclude,
    verbose=args.verbose)

if not RELATIVE_DIRECTORY_TO_DATA_FILES_LIST:
    p.error(f"No images or data files found in {args.directory}")


# parse metadata from flipbook_metadata.json files
METADATA_COLUMNS, RELATIVE_DIRECTORY_TO_METADATA = get_relative_directory_to_metadata(
    args.directory,
    RELATIVE_DIRECTORY_TO_DATA_FILES_LIST,
    verbose=args.verbose)


# parse metadata from the metadata_table if specified
if args.metadata_table and os.path.isfile(args.metadata_table):
    #args.metadata_table = os.path.join(args.directory, args.metadata_table)
    try:
        df = parse_table(args.metadata_table)
    except ValueError as e:
        p.error(str(e))

    for relative_directory, row in df.iterrows():
        metadata_dict = {k: v for k, v in row.to_dict().items() if k != PATH_COLUMN}
        if relative_directory not in RELATIVE_DIRECTORY_TO_METADATA:
            RELATIVE_DIRECTORY_TO_METADATA[relative_directory] = {}
            if args.verbose:
                print(f"Setting new metadata row for {relative_directory} to {metadata_dict}")
        else:
            if args.verbose:
                print(f"Updating existing metadata row for {relative_directory} to {metadata_dict}")
        RELATIVE_DIRECTORY_TO_METADATA[relative_directory].update(metadata_dict)

        for key in metadata_dict:
            if key not in METADATA_COLUMNS:
                METADATA_COLUMNS.append(key)

# define input form fields to show on each data page
FORM_SCHEMA = [
    {
        "type": "radio",
        "columnName": "Verdict",
        "choices": [
            {"value": "normal", "label": "Normal"},
            {"value": "intermediate", "label": "Intermediate"},
            {"value": "full-expansion", "label": "Full Expansion"},
            {"value": "double-expansion", "label": "Double Expansion"},
        ]
    },
    {
        "type": "radio",
        "columnName": "Confidence",
        "inputLabel": "Confidence",
        "choices": [
            {"value": "borderline", "label": "Borderline"},
            {"value": "confident", "label": "Confident"},
        ]
    },
    {
        "type": "text",
        "columnName": "Notes",
        "size": 60
    }
]

if args.form_schema_json:
    print(f"Loading form schema from {args.form_schema_json}")
    try:
        if os.path.isfile(args.form_schema_json):
            with open(args.form_schema_json, "rt") as f:
                FORM_SCHEMA = json.load(f)
        elif args.form_schema_json.startswith("http"):
            # Convert https://github.com/broadinstitute/flipbook/blob/main/flipbook/__init__.py to the raw url:
            # https://raw.githubusercontent.com/broadinstitute/flipbook/main/flipbook/__init__.py
            github_match = re.search("//github.com/(.*)/blob/(.*)", args.form_schema_json)
            if github_match:
                part1 = github_match.group(1)
                part2 = github_match.group(2)
                github_raw_file_url = f"https://raw.githubusercontent.com/{part1}/{part2}"
                if args.verbose:
                    print(f"Converting form schema url {args.form_schema_json} to {github_raw_file_url}")
                args.form_schema_json = github_raw_file_url
            r = requests.get(url=args.form_schema_json)
            FORM_SCHEMA = r.json()
    except Exception as e:
        p.error(f"Couldn't parse {args.form_schema_json}: {e}")


FORM_SCHEMA_COLUMNS = [r['columnName'] for r in FORM_SCHEMA]

# validate FORM_SCHEMA and determine keyboard shortcuts
FORM_RADIO_BUTTON_KEYBOARD_SHORTCUTS = {}
for i, form_schema_row in enumerate(FORM_SCHEMA):
    if not isinstance(form_schema_row, dict):
        raise ValueError(f"FORM_SCHEMA row {i} must be a dictionary")

    missing_keys = {'type', 'columnName'} - set(form_schema_row.keys())
    if missing_keys:
        raise ValueError(f"FORM_SCHEMA row {i} is missing values for these keys: {', '.join(missing_keys)}")
    if form_schema_row['type'] not in ('text', 'radio'):
        raise ValueError(f"FORM_SCHEMA row {i} has unexpected 'type' value: {form_schema_row['type']}")
    if 'name' not in form_schema_row:
        form_schema_row['name'] = form_schema_row['columnName'].lower()
    form_schema_row['name'] = re.sub("[^a-zA-Z0-9_]", "_", form_schema_row['name'])
    if 'inputLabel' not in form_schema_row:
        form_schema_row['inputLabel'] = form_schema_row['columnName']

    if form_schema_row['type'] == 'radio':
        if 'choices' not in form_schema_row or not isinstance(form_schema_row['choices'], list):
            raise ValueError(f"FORM_SCHEMA row {i} is missing a 'choices' list")
        for choice in form_schema_row['choices']:
            if not isinstance(choice, dict):
                raise ValueError(f"FORM_SCHEMA row {i} must 'choices' list must contain dictionaries")
            missing_keys = {'value', 'label'} - set(choice.keys())
            if missing_keys:
                raise ValueError(f"FORM_SCHEMA row {i} 'choices' list has entries where these keys are missing: {', '.join(missing_keys)}")

            label_without_html = re.sub("<[^<]+?>", "", choice['label']).strip()
            first_letter = (label_without_html or choice["value"])[0]
            FORM_RADIO_BUTTON_KEYBOARD_SHORTCUTS[first_letter] = choice['value']
            print(f"Form Keyboard Shortcut: {first_letter} => {choice['label']}")

# parse or create FORM_RESPONSES dict for storing user responses
FORM_RESPONSES = {}

# if a form_responses_table is provided with additional columns which are not in the current FORM_SCHEMA (eg. if the
# form schema changes, save this info here so it's not lost when the table is updated after new form responses.
EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE = []
EXTRA_DATA_IN_FORM_RESPONSES_TABLE = {}

if FORM_SCHEMA:
    args.form_responses_table = os.path.join(args.directory, args.form_responses_table)
    args.form_responses_table_is_excel = is_excel_table(args.form_responses_table)

    if os.path.isfile(args.form_responses_table):
        try:
            df = parse_table(args.form_responses_table)
        except ValueError as e:
            p.error(str(e))
        EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE = [c for c in df.columns if c not in FORM_SCHEMA_COLUMNS and c not in METADATA_COLUMNS and c != PATH_COLUMN]
    else:
        # make sure the table can be written out later
        if not os.access(os.path.dirname(args.form_responses_table), os.W_OK):
            p.error(f"Unable to create {args.form_responses_table}")

        # create empty table in memory
        df = pd.DataFrame(columns=[PATH_COLUMN] + FORM_SCHEMA_COLUMNS)
        df.set_index(PATH_COLUMN, inplace=True, drop=False)

    # store form responses in memory.
    # NOTE: This is not thread-safe and assumes a single-threaded server. For multi-threaded
    # or multi-process servers like gunicorn, this will need to be replaced with a sqlite or redis backend.
    FORM_RESPONSES = collections.OrderedDict()
    for relative_directory, row in df.iterrows():
        row_as_dict = row.to_dict()
        FORM_RESPONSES[relative_directory] = {k: v for k, v in row_as_dict.items() if k in FORM_SCHEMA_COLUMNS}
        EXTRA_DATA_IN_FORM_RESPONSES_TABLE[relative_directory] = {k: v for k, v in row_as_dict.items() if k not in FORM_SCHEMA_COLUMNS and k != PATH_COLUMN}

    print(f"Will save form responses to {args.form_responses_table}  (columns: {', '.join(FORM_SCHEMA_COLUMNS + EXTRA_COLUMNS_IN_FORM_RESPONSES_TABLE)})")


if args.sort_by:
    valid_columns = set([PATH_COLUMN])
    valid_columns.update(FORM_SCHEMA_COLUMNS)
    if METADATA_COLUMNS:
        valid_columns.update(METADATA_COLUMNS)

    invalid_values = ", ".join([f"'{s}'" for s in args.sort_by if s not in valid_columns])
    if invalid_values:
        p.error(f"{invalid_values} column(s) not found in metadata. --sort-by value should be one of: " +
                ", ".join(valid_columns))

    print(f"Sorting {len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)} pages by {', '.join(args.sort_by)}")
    def get_sort_key(i):
        relative_dir = i[0]
        sort_key = []
        for s in args.sort_by:
            if s == PATH_COLUMN:
                sort_key.append(relative_dir)
                continue
            form_value = FORM_RESPONSES.get(relative_dir, {}).get(s)
            if form_value is not None:
                sort_key.append(str(form_value))
                continue
            metadata_value = RELATIVE_DIRECTORY_TO_METADATA.get(relative_dir, {}).get(s)
            if metadata_value is not None:
                sort_key.append(str(metadata_value))
                continue
        return tuple(sort_key)

    RELATIVE_DIRECTORY_TO_DATA_FILES_LIST = sorted(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST, key=get_sort_key)


def send_file(path):
    print(f"Sending {args.directory} {path}")
    if path.startswith("favicon"):
        return Response(pkg_resources.resource_stream('flipbook', 'icons/favicon.png'), mimetype='image/png')

    return send_from_directory(args.directory, path, as_attachment=True)


def get_static_data_page_url(page_number, last):
    i = page_number - 1
    if i < 0 or i >= len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST):
        raise ValueError(f"page_number arg is out of bounds. It must be between 1 and {len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)}")

    relative_dir, _ = RELATIVE_DIRECTORY_TO_DATA_FILES_LIST[i]
    name = relative_dir.replace("/", "__")
    return f"page_{name}.html"


def main():
    # add a ctime(..) function to allow the last-changed-time of a path to be computed within a jinja template
    jinja2.environment.DEFAULT_FILTERS['ctime'] = lambda path: int(os.path.getctime(path)) if os.path.isfile(path) else 0

    from flipbook.main_list import main_list_handler
    from flipbook.data_page import data_page_handler
    from flipbook.save import save_form_handler

    app = Flask(__name__)

    if args.generate_static_website:
        os.makedirs(WEBSITE_DIR, exist_ok=True)

        with open(os.path.join(WEBSITE_DIR, "index.html"), "wt") as f:
            f.write(main_list_handler(is_static_website=True).get_data(as_text=True))
        with open(os.path.join(WEBSITE_DIR, "favicon.png"), "wb") as f:
            f.write(pkgutil.get_data('flipbook', 'icons/favicon.png'))

        last_page_number = len(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST)
        for i, (relative_directory, data_file_types_and_paths) in enumerate(RELATIVE_DIRECTORY_TO_DATA_FILES_LIST):
            page_number = i + 1
            with app.test_request_context(get_data_page_url(page_number, last_page_number)):
                page_dir = os.path.join(WEBSITE_DIR, relative_directory)
                os.makedirs(page_dir, exist_ok=True)
                for data_file_type, data_file in data_file_types_and_paths:
                    if data_file_type in (METADATA_JSON_FILE_TYPE, CONTENT_HTML_FILE_TYPE):
                        continue
                    print("Copying", data_file_type, data_file, "to", page_dir)
                    shutil.copy(data_file, page_dir)

                with open(os.path.join(WEBSITE_DIR, get_static_data_page_url(page_number, last_page_number)), "wt") as f:
                    f.write(data_page_handler(is_static_website=True).get_data(as_text=True))
        print("Done")
        print(f"Generated static website in the ./{WEBSITE_DIR} directory")
        sys.exit(0)

    # start web server
    CORS(app)

    app.url_map.strict_slashes = False

    app.add_url_rule('/', view_func=main_list_handler, methods=['GET'])
    app.add_url_rule('/page', view_func=data_page_handler, methods=['POST', 'GET'])
    app.add_url_rule('/save', view_func=save_form_handler, methods=['POST'])
    app.add_url_rule('/<path:path>', view_func=send_file, methods=['GET'])

    os.environ["WERKZEUG_RUN_MAIN"] = "true"

    host = os.environ.get('HOST', args.host)
    port = int(os.environ.get('PORT', args.port))
    if args.verbose:
        print(f"Connecting to {host}:{port}")

    app.run(
        debug=args.dev_mode,
        host=host,
        port=port)


import os
import sys
from setuptools import setup

command = sys.argv[-1]
if command == 'publish':
    os.system('rm -rf dist')
    os.system('python3 setup.py sdist')
    os.system('python3 setup.py bdist_wheel')
    os.system('twine upload dist/*whl dist/*gz')
    sys.exit()

with open("README.md", "rt") as fh:
    long_description = fh.read()

install_requires = [
    "configargparse",
    "flask",
    "flask-cors",
    "gunicorn",
    "jinja2",
    "openpyxl",
    "pandas",
    "requests",
    "wcmatch",
    "xlrd",
    "xlwt",
]

setup(
    name='flipbook',
    version="0.11.1",
    description="Starts a simple image server that lets you quickly flip through image files from a local directory "
                "using your web browser and optionally answering customizable questions about each one",
    install_requires=install_requires,
    entry_points = {
        'console_scripts': [
            'flipbook = flipbook:main',
        ],
    },
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=["flipbook"],
    py_modules=["compare_form_response_tables"],
    include_package_data=True,
    package_data={'': ['*/*.png', '*/*.html']},
    python_requires=">=3.7",
    license="MIT",
    keywords='curation, NGS, sequencing, STRs, REviewer, read visualization, machine learning',
    url='https://github.com/broadinstitute/flipbook',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)
